/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDEING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */
/* dw_print.c - format dreamweaver lock tags for different xssi
 * directives.  a couple handly string manipulation routines too.
 *
 * original by alxander ladygo <alx@taylor.org>.
 */

#include "httpd.h"
#if defined(WIN32)
#include "../os/win32/os.h"
#elif defined(macintosh) 
#include "os.h"
#endif
#include "http_protocol.h"
#include "dw_print.h"
#include "../xs.h"

/*
 * create the locking tags around the translated directive, and
 * print to the buffer.
 */

char *dw_cur_tag_str[5] = {
	"<!-- empty -->",
	DW_IF_XML,
	DW_ELIF_XML,
	DW_ELSE_XML,
	DW_ENDIF_XML
};

char *dw_current_tag_str(int i) {
	return dw_cur_tag_str[i];
}

void dw_print_echo(request_rec *r, const char *tag, const char *val, int count)
{
	char header[MAX_LOCK_LEN], orig[MAX_LOCK_LEN];

	sprintf(orig, DW_ECHO_ORIG, dw_escape_quotes(r->pool, tag));
	sprintf(header, DW_LOCK_BEGIN_STR, count, DW_ECHO_TYPE,
					dw_mini_escape(r->pool, orig));
	ap_rputs(header, r);
	ap_rputs(val, r);
	ap_rputs(DW_LOCK_END_STR, r);
}

void dw_print_set(request_rec *r, const char *tag,
									const char *val, const char *val_orig, int count)
{
	char header[MAX_LOCK_LEN], orig[MAX_LOCK_LEN];

	sprintf(orig, DW_SET_ORIG, dw_escape_quotes(r->pool, tag),
					dw_escape_quotes(r->pool, val_orig));
	sprintf(header, DW_LOCK_BEGIN_STR, count, DW_SET_TYPE,
					dw_mini_escape(r->pool, orig));
	ap_rputs(header, r);
	ap_rputs(DW_SET_XML, r);
	ap_rputs(DW_LOCK_END_STR, r);
}

void dw_print_printenv(request_rec *r, const char *env, int count)
{
	char header[MAX_LOCK_LEN];

	sprintf(header, DW_LOCK_BEGIN_STR, count, DW_PRINTENV_TYPE,
					dw_mini_escape(r->pool, DW_PRINTENV_ORIG));
	ap_rputs(header, r);
	ap_rputs(env, r);
	ap_rputs(DW_LOCK_END_STR, r);
}

void dw_print_vinclude_head(request_rec *r, const char *fname, int count)
{
	char header[MAX_LOCK_LEN], orig[MAX_LOCK_LEN];

	sprintf(orig, DW_VINCLUDE_ORIG, dw_escape_quotes(r->pool, fname));
	sprintf(header, DW_LOCK_BEGIN_STR, count, DW_INCLUDE_TYPE,
					dw_mini_escape(r->pool, orig));
	ap_rputs(header, r);
}

void dw_print_finclude_head(request_rec *r, const char *fname, int count)
{
	char header[MAX_LOCK_LEN], orig[MAX_LOCK_LEN];

	sprintf(orig, DW_FINCLUDE_ORIG, dw_escape_quotes(r->pool, fname));
	sprintf(header, DW_LOCK_BEGIN_STR, count, DW_INCLUDE_TYPE,
					dw_mini_escape(r->pool, orig));
	ap_rputs(header, r);
}

void dw_print_include_tail(request_rec *r)
{
	ap_rputs(DW_LOCK_END_STR, r);
}

void dw_print_if(request_rec *r, const char *expr, int printing,
								 int current_tag, int count)
{
	char header[MAX_LOCK_LEN], orig[MAX_LOCK_LEN];

	sprintf(orig, DW_IF_ORIG, dw_escape_quotes(r->pool, expr));

	if (printing) {
		sprintf(header, DW_IF_LOCK_BEGIN_STR, count, DW_IF_TYPE, printing,
						dw_mini_escape(r->pool, orig));
		ap_rputs(header, r);
		/*sprintf(invis, DW_IF_XML, printing);*/
		ap_rputs(DW_IF_XML, r);
		/*ap_rputs(orig, r);*/
		ap_rputs(DW_LOCK_END_STR, r);
	} else {
		sprintf(header, DW_SEMI_IF_LOCK_BEGIN_STR, count, DW_IF_TYPE, printing,
						dw_mini_escape(r->pool, orig));
		ap_rputs(header, r);
	}
}

void dw_print_elif(request_rec *r, const char *expr, int printing,
									 int just_printed, int current_tag,
									 dw_strbuff *savebuff, int count)
{
	char header[MAX_LOCK_LEN], orig[MAX_LOCK_LEN];

	sprintf(orig, DW_ELIF_ORIG, dw_escape_quotes(r->pool, expr));

	/* if just_printed is nonzero, the previous "if" or "elif" printed */
	if (!just_printed && savebuff->outdata != NULL) {
		/* write the savebuff things, and end the previous 'if' locking region */
		ap_rputs(dw_mini_escape(r->pool, savebuff->outdata), r);
		savebuff->outdata = NULL;
		/* XXX clean this up, print status to DW_IF_INVIS */
		sprintf(header, DW_SEMI_LOCK_END_STR,
						dw_current_tag_str(current_tag));
		ap_rputs(header, r);
	}

	if (printing) {
		/* write new full lock */
		sprintf(header, DW_IF_LOCK_BEGIN_STR, count, DW_ELIF_TYPE, printing,
						dw_mini_escape(r->pool, orig));
		ap_rputs(header, r);
		ap_rputs(DW_ELIF_XML, r);
		ap_rputs(DW_LOCK_END_STR, r);
	} else {
		/* write the semi-lock */
		sprintf(header, DW_SEMI_IF_LOCK_BEGIN_STR, count, DW_ELIF_TYPE, printing,
						dw_mini_escape(r->pool, orig));
		ap_rputs(header, r);
	}
}

void dw_print_else(request_rec *r, int printing, int just_printed,
									 int current_tag, dw_strbuff *savebuff, int count)
{
	char header[MAX_LOCK_LEN];

	if (!just_printed && savebuff->outdata != NULL) {
		/* write the savebuff things, and end the previous 'if' locking region */
		ap_rputs(dw_mini_escape(r->pool, savebuff->outdata), r);
		savebuff->outdata = NULL;
		sprintf(header, DW_SEMI_LOCK_END_STR, dw_current_tag_str(current_tag));
		ap_rputs(header, r);
	}
	
	/* assumes the value of "printing" is for the current else directive. */
	if (printing) {
		/* write new full lock */
		sprintf(header, DW_IF_LOCK_BEGIN_STR, count, DW_ELSE_TYPE, printing,
						dw_mini_escape(r->pool, DW_ELSE_ORIG));
		ap_rputs(header, r);
		ap_rputs(DW_ELSE_XML, r);
		ap_rputs(DW_LOCK_END_STR, r);
	} else {
		/* write the semi-lock */
		sprintf(header, DW_SEMI_IF_LOCK_BEGIN_STR, count, DW_ELSE_TYPE, printing,
						dw_mini_escape(r->pool, DW_ELSE_ORIG));
		ap_rputs(header, r);
	}
}

void dw_print_endif(request_rec *r, int printing, int current_tag,
										dw_strbuff *savebuff, int count)
{
	char header[MAX_LOCK_LEN];

	/* assumes the value of "printing" hasn't changed since the last
	 * directive.  opposite of "else" above. -alx */
	if (!printing) {
		/* write the savebuff things, and end the previous 'if' locking region */
		ap_rputs(dw_mini_escape(r->pool, savebuff->outdata), r);
		savebuff->outdata = NULL;
		sprintf(header, DW_SEMI_LOCK_END_STR, dw_current_tag_str(current_tag));
		ap_rputs(header, r);
	}
	
	sprintf(header, DW_LOCK_BEGIN_STR, count, DW_ENDIF_TYPE,
					dw_mini_escape(r->pool, DW_ENDIF_ORIG));
	ap_rputs(header, r);
	ap_rputs(DW_ENDIF_XML, r);
	ap_rputs(DW_LOCK_END_STR, r);
}


/* turn " into \" */
char *dw_escape_quotes(pool *pool, const char *orig)
{
	char *newstr, *newptr;
	int i, len, num_converts = 0;
	
	len = strlen(orig);
	for (i = 0; i < len; ++i) {
		if (orig[i] == '"')
			++num_converts;
	}

	if ((newstr = ap_palloc(pool, len + num_converts + 1)) == NULL) {
		/* XXX: handle error */
		return "argh";
	}
	newptr = newstr;

	while (*orig != '\0') {
		if (*orig == '"') {
			strcpy(newptr, "\\\"");
			newptr += 2;
			orig++;
		} else {
			*newptr = *orig;
			newptr++;
			orig++;
		}
	}
	*newptr = '\0';

	return newstr;
}

/* minimal escaping as per dreamweaver automated translation doc. */
char *dw_mini_escape(pool *pool, const char *orig)
{
	/* i don't really care that this may be slow.  orig is likely to
	 * be under 100 bytes anyway. -alx */
	char *newstr, *newptr;
	int i, len, num_converts = 0;
	
	len = strlen(orig);
	for (i = 0; i < len; ++i) {
		switch ((int)orig[i]) {
		case '"':
		case '<':
		case '>':
			++num_converts;
			break;
		}
	}

	if ((newstr = ap_palloc(pool, len + num_converts * 2 + 1)) == NULL) {
		/* XXX: handle error */
		return "argh";
	}
	newptr = newstr;

	while (*orig != '\0') {
		switch ((int)*orig) {
		case '"':
			strcpy(newptr, "%22");
			newptr += 3;
			orig++;
			break;
		case '<':
			strcpy(newptr, "%3C");
			newptr += 3;
			orig++;
			break;
		case '>':
			strcpy(newptr, "%3E");
			newptr += 3;
			orig++;
			break;
		default:
			*newptr = *orig;
			newptr++;
			orig++;
			break;
		}
	}
	*newptr = '\0';

	return newstr;
}
