/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDEING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */
/* dw_strbuff.c - your basic growable string implementation.
 *
 * original by alxander ladygo <alx@taylor.org>.
 */

#include <string.h>
#include <math.h>
#include "dw_strbuff.h"
#include "alloc.h"

void dw_strwritec(dw_strbuff *sb, char c)
{
	char out[2];
	out[0] = c;
	out[1] = '\0';
	dw_strwrite(sb, out, 1);
}

void dw_strwrite(dw_strbuff *sb, const char *out, int nbyte) {
	char *tmpdata, *newdata;
	int ptrdist, newsize, freesize, numblocks;
	
	if (sb->outdata == NULL) {
		sb->buffsize = DEFAULT_BUFSIZE;
		sb->outdata = (char *)ap_palloc(sb->pool, sizeof(char) * sb->buffsize);
		sb->outdata[0] = '\0';
		sb->outptr = sb->outdata;
	}
	
	freesize = sb->buffsize - strlen(sb->outdata) - 1; /* one for the '\0' */

	if (nbyte > freesize) {
		ptrdist = sb->outptr - sb->outdata;

#ifdef macintosh
		numblocks = (int)ceilf((float)(nbyte - freesize) / (float)DEFAULT_BUFSIZE);
#else
		numblocks = (int)ceil((float)(nbyte - freesize) / (float)DEFAULT_BUFSIZE);
#endif
		newsize = sb->buffsize + (numblocks * DEFAULT_BUFSIZE);

		newdata = (char *)ap_palloc(sb->pool, sizeof(char) * newsize);
		memcpy(newdata, sb->outdata, sb->buffsize);
		tmpdata = sb->outdata;
		sb->outdata = newdata;
		sb->outptr = newdata + ptrdist;
		sb->buffsize = newsize;
		/*free(tmpdata); in truth, this is handled my apache pools, right? -alx */
	}

	memcpy(sb->outptr, out, nbyte);
	sb->outptr += nbyte;
	*(sb->outptr) = '\0';

	/*printf("sizeof outdata len: %d\n", strlen(sb->outdata));*/
	/*printf("so far: \"%s\"\n", sb->outdata);*/
}
