/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDEING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */
/* enviro.c - stuff to set up HTTP environmental variables.
 * many of these are defined and passed in by the xssi spoofing
 * functions of the javascript extensions, and are therefore
 * commented out here.  that's probably a better way to add new
 * variables anyway.
 *
 * original by alxander ladygo <alx@taylor.org>.
 */

#include <string.h>
#include "ap_config.h"
#include "ap_compat.h"
#include "hsregex.h"
#include "ap.h"
#include "httpd.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_config.h"
#include "ap_mmn.h"


char *dw_default_type() {
    return "text/html";
}

/*
 * this is all stuff that will need to be filled out
 * once we have a real environment generator.
 */

static char *dw_get_server_version() {
    return "Apache/1.3.3-spoof";
}

char *dw_get_server_name(request_rec *r) {
    return "www.nowhere.org";
}

static unsigned dw_get_server_port(request_rec *r) {
    return 80U;
}

static char *dw_get_remote_host(conn_rec *c, request_rec *r) {
    return "machine.nowhere.org";
}

static char *dw_get_remote_addr(conn_rec *c, request_rec *r) {
    return "192.168.0.1";
}

static char *dw_document_root(request_rec *r) {
    return r->server->ap_document_root;
}

static char *dw_server_admin() {
    return "webmaster@nowhere.org";
}

static unsigned dw_get_remote_port(conn_rec *c, request_rec *r) {
    return 1995;
}

static char *dw_user() {
    return "xebeche";
}

static char *dw_auth_type() {
    return "Basic";
}

static char *dw_get_remote_logname() {
    return "nobody";
}

static char *dw_protocol() {
    return "HTTP/1.0";
}

static char *dw_original_uri(request_rec *r) {
    return r->uri;
}

/* stolen and hacked from http_core */
API_EXPORT(const char *) dw_psignature(const char *prefix, request_rec *r)
{
    char sport[20];

    ap_snprintf(sport, sizeof sport, "%u", (unsigned) dw_get_server_port(r));

    return ap_pstrcat(r->pool, prefix, "<ADDRESS>" SERVER_BASEVERSION
		      " Server at ", dw_get_server_name(r), " Port ", sport,
		      "</ADDRESS>\n", NULL);
}

/* originally stolen from util_script.c, but this is specific enough
 * to the dreamweaver situation that it warrants a separate file. -alx
 */

static char *http2env(pool *a, char *w)
{
    char *res = ap_pstrcat(a, "HTTP_", w, NULL);
    char *cp = res;

    while (*++cp) {
	if (!ap_isalnum(*cp) && *cp != '_') {
	    *cp = '_';
	}
	else {
	    *cp = ap_toupper(*cp);
	}
    }

    return res;
}

/* This "cute" little function comes about because the path info on
 * filenames and URLs aren't always the same. So we take the two,
 * and find as much of the two that match as possible.
 */

API_EXPORT(int) ap_find_path_info(const char *uri, const char *path_info)
{
    int lu = strlen(uri);
    int lp = strlen(path_info);

    while (lu-- && lp-- && uri[lu] == path_info[lp]);

    if (lu == -1) {
	lu = 0;
    }

    while (uri[lu] != '\0' && uri[lu] != '/') {
        lu++;
    }
    return lu;
}

API_EXPORT(void) ap_add_common_vars(request_rec *r)
{
	table *e;
	/*server_rec *s = r->server;*/
	/*conn_rec *c = r->connection;*/
	/*const char *rem_logname;*/
	char *env_path;
#ifdef WIN32
	char *env_temp;
#endif
	/*const char *host;*/
	array_header *hdrs_arr = ap_table_elts(r->headers_in);
	table_entry *hdrs = (table_entry *) hdrs_arr->elts;
	/*int i;*/

	/* use a temporary table which we'll overlap onto
	 * r->subprocess_env later
	 */
	e = ap_make_table(r->pool, 25 + hdrs_arr->nelts);

	/* First, add environment vars from headers... this is as per
	 * CGI specs, though other sorts of scripting interfaces see
	 * the same vars...
	 */

	/* -no headers, eh?
	for (i = 0; i < hdrs_arr->nelts; ++i) {
		if (!hdrs[i].key) {
	    continue;
		}
	*/
		/* A few headers are special cased --- Authorization to prevent
		 * rogue scripts from capturing passwords; content-type and -length
		 * for no particular reason.
		 */
		/*
			if (!strcasecmp(hdrs[i].key, "Content-type")) {
	    ap_table_addn(e, "CONTENT_TYPE", hdrs[i].val);
			}
			else if (!strcasecmp(hdrs[i].key, "Content-length")) {
	    ap_table_addn(e, "CONTENT_LENGTH", hdrs[i].val);
			}
		*/
		/*
		 * You really don't want to disable this check, since it leaves you
		 * wide open to CGIs stealing passwords and people viewing them
		 * in the environment with "ps -e".  But, if you must...
		 */
		/*
			#ifndef SECURITY_HOLE_PASS_AUTHORIZATION
			else if (!strcasecmp(hdrs[i].key, "Authorization") 
			|| !strcasecmp(hdrs[i].key, "Proxy-Authorization")) {
	    continue;
			}
			#endif
			else {
	    ap_table_addn(e, http2env(r->pool, hdrs[i].key), hdrs[i].val);
			}
			}
		*/

	if (!(env_path = getenv("PATH"))) {
		env_path = DEFAULT_PATH;
	}

#ifdef WIN32
	if (env_temp = getenv("SystemRoot")) {
		ap_table_addn(e, "SystemRoot", env_temp);         
	}
	if (env_temp = getenv("COMSPEC")) {
		ap_table_addn(e, "COMSPEC", env_temp);            
	}
	if (env_temp = getenv("WINDIR")) {
		ap_table_addn(e, "WINDIR", env_temp);
	}
#endif

	ap_table_addn(e, "PATH", env_path);
	/*
	ap_table_setn(e, "SERVER_SIGNATURE", dw_psignature("", r));
	ap_table_addn(e, "SERVER_SOFTWARE", dw_get_server_version());
	ap_table_addn(e, "SERVER_NAME", dw_get_server_name(r));
	ap_table_addn(e, "SERVER_PORT",
								ap_psprintf(r->pool, "%u", dw_get_server_port(r)));
	host = dw_get_remote_host(c, r);
	if (host) {
		ap_table_addn(e, "REMOTE_HOST", host);
	}
	ap_table_addn(e, "REMOTE_ADDR", dw_get_remote_addr(c, r));
	*/
	ap_table_addn(e, "DOCUMENT_ROOT", dw_document_root(r));	/* Apache */
	
	/*ap_table_addn(e, "SERVER_ADMIN", dw_server_admin());	/* Apache */

	ap_table_addn(e, "SCRIPT_FILENAME", r->filename);	/* Apache */
	/*
	ap_table_addn(e, "REMOTE_PORT",
								ap_psprintf(r->pool, "%d", dw_get_remote_port(c, r)));

	ap_table_addn(e, "REMOTE_USER", dw_user());
	ap_table_addn(e, "AUTH_TYPE", dw_auth_type());

	rem_logname = dw_get_remote_logname(r);
	if (rem_logname) {
		ap_table_addn(e, "REMOTE_IDENT", ap_pstrdup(r->pool, rem_logname));
	}
	*/
	/* Apache custom error responses. If we have redirected set two new vars */
	/* unused, no redirects in dreamweaver -alx
		 if (r->prev) {
		 if (r->prev->args) {
		 ap_table_addn(e, "REDIRECT_QUERY_STRING", r->prev->args);
		 }
		 if (r->prev->uri) {
		 ap_table_addn(e, "REDIRECT_URL", r->prev->uri);
		 }
		 }
	*/

	ap_overlap_tables(r->subprocess_env, e, AP_OVERLAP_TABLES_SET);
}

API_EXPORT(void) ap_add_cgi_vars(request_rec *r)
{
	table *e = r->subprocess_env;
	/*
	ap_table_setn(e, "GATEWAY_INTERFACE", "CGI/1.1");
	ap_table_setn(e, "SERVER_PROTOCOL", dw_protocol());
	ap_table_setn(e, "REQUEST_METHOD", r->method);
	*/
	ap_table_setn(e, "QUERY_STRING", r->args ? r->args : "");
	ap_table_setn(e, "REQUEST_URI", dw_original_uri(r));

	/* Note that the code below special-cases scripts run from includes,
	 * because it "knows" that the sub_request has been hacked to have the
	 * args and path_info of the original request, and not any that may have
	 * come with the script URI in the include command.  Ugh.
	 */

	if (!strcmp(dw_protocol(), "INCLUDED")) {
		ap_table_setn(e, "SCRIPT_NAME", r->uri);
		if (r->path_info && *r->path_info) {
	    ap_table_setn(e, "PATH_INFO", r->path_info);
		}
	}
	else if (!r->path_info || !*r->path_info) {
		ap_table_setn(e, "SCRIPT_NAME", r->uri);
	}
	else {
		int path_info_start = ap_find_path_info(r->uri, r->path_info);

		ap_table_setn(e, "SCRIPT_NAME",
									ap_pstrndup(r->pool, r->uri, path_info_start));

		ap_table_setn(e, "PATH_INFO", r->path_info);
	}

	if (r->path_info && r->path_info[0]) {
		/*
		 * To get PATH_TRANSLATED, treat PATH_INFO as a URI path.
		 * Need to re-escape it for this, since the entire URI was
		 * un-escaped before we determined where the PATH_INFO began.
		 */
		request_rec *pa_req;

		pa_req = ap_sub_req_lookup_uri(escape_uri(r->pool, r->path_info), r);

		if (pa_req->filename) {
#ifdef WIN32
	    char buffer[HUGE_STRING_LEN];
#endif
	    char *pt = ap_pstrcat(r->pool, pa_req->filename, pa_req->path_info,
														NULL);
#ifdef WIN32
	    /* We need to make this a real Windows path name */
	    GetFullPathName(pt, HUGE_STRING_LEN, buffer, NULL);
	    ap_table_setn(e, "PATH_TRANSLATED", ap_pstrdup(r->pool, buffer));
#else
	    ap_table_setn(e, "PATH_TRANSLATED", pt);
#endif
		}
		ap_destroy_sub_req(pa_req);
	}
}

