/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * http_log.c: Dealing with the logs and errors
 * 
 * Rob McCool
 * 
 */


#define CORE_PRIVATE
#include <stdarg.h>
#include "httpd.h"
/*
#include "http_config.h"
#include "http_core.h"
#include "http_main.h"
*/
#include "http_log.h"
#include "ap.h"
#include "../xs.h"
#include "dw_strbuff.h"

typedef struct {
	char	*t_name;
	int	t_val;
} TRANS;

static const TRANS priorities[] = {
    {"emerg",	APLOG_EMERG},
    {"alert",	APLOG_ALERT},
    {"crit",	APLOG_CRIT},
    {"error",	APLOG_ERR},
    {"warn",	APLOG_WARNING},
    {"notice",	APLOG_NOTICE},
    {"info",	APLOG_INFO},
    {"debug",	APLOG_DEBUG},
    {NULL,	-1},
};

extern dw_strbuff *error_buff;
/*
API_EXPORT(void) ap_error_log2stderr(server_rec *s) {
	if (   s->error_log != NULL
				 && fileno(s->error_log) != STDERR_FILENO)
		dup2(fileno(s->error_log), STDERR_FILENO);
}
*/
static void log_error_core(const char *file, int line, int level,
													 const server_rec *s, const request_rec *r,
													 const char *fmt, va_list args)
{
	char errstr[MAX_STRING_LEN];
	size_t len = 0;
	int save_errno = errno;

	/* this was gutted and replaced with a write to error_buff. -alx */

	len += ap_snprintf(errstr + len, sizeof(errstr) - len,
										 "[%s] ", priorities[level & APLOG_LEVELMASK].t_name);

	if (file && (level & APLOG_LEVELMASK) == APLOG_DEBUG) {
		len += ap_snprintf(errstr + len, sizeof(errstr) - len,
											 "%s(%d): ", file, line);
	}

	if (!(level & APLOG_NOERRNO)
			&& (save_errno != 0)
#ifdef WIN32
			&& !(level & APLOG_WIN32ERROR)
#endif
			) {
		len += ap_snprintf(errstr + len, sizeof(errstr) - len,
											 "(%d)%s: ", save_errno, strerror(save_errno));
	}
#ifdef WIN32
	if (level & APLOG_WIN32ERROR) {
		int nChars;
		int nErrorCode;

		nErrorCode = GetLastError();
		len += ap_snprintf(errstr + len, sizeof(errstr) - len,
											 "(%d)", nErrorCode);

		nChars = FormatMessage( 
													 FORMAT_MESSAGE_FROM_SYSTEM,
													 NULL,
													 nErrorCode,
													 MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
													 (LPTSTR) errstr + len,
													 sizeof(errstr) - len,
													 NULL 
													 );
		len += nChars;
		if (nChars == 0) {
	    /* Um, error occurred, but we can't recurse to log it again
	     * (and it would probably only fail anyway), so lets just
	     * log the numeric value.
	     */
	    nErrorCode = GetLastError();
	    len += ap_snprintf(errstr + len, sizeof(errstr) - len,
												 "(FormatMessage failed with code %d): ", nErrorCode);
		}
		else {
	    /* FormatMessage put the message in the buffer, but it may
	     * have appended a newline (\r\n). So remove it and use
	     * ": " instead like the Unix errors. The error may also
	     * end with a . before the return - if so, trash it.
	     */
	    if (len > 1 && errstr[len-2] == '\r' && errstr[len-1] == '\n') {
				if (len > 2 && errstr[len-3] == '.')
					len--;
				errstr[len-2] = ':';
				errstr[len-1] = ' ';
	    }
		}
	}
#endif

	len += ap_vsnprintf(errstr + len, sizeof(errstr) - len, fmt, args);

	dw_strwrite(error_buff, errstr, strlen(errstr));
	dw_strwritec(error_buff, '\n');
}

API_EXPORT(void) ap_log_error(const char *file, int line, int level,
			      const server_rec *s, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	log_error_core(file, line, level, s, NULL, fmt, args);
	va_end(args);
}

API_EXPORT(void) ap_log_rerror(const char *file, int line, int level,
			       const request_rec *r, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	/*log_error_core(file, line, level, r->server, r, fmt, args);  no server, eh? -alx */
	log_error_core(file, line, level, NULL, r, fmt, args);
	/*
	 * IF the error level is 'warning' or more severe,
	 * AND there isn't already error text associated with this request,
	 * THEN make the message text available to ErrorDocument and
	 * other error processors.  This can be disabled by stuffing
	 * something, even an empty string, into the "error-notes" cell
	 * before calling this routine.
	 */
	if (((level & APLOG_LEVELMASK) <= APLOG_WARNING)
			&& (ap_table_get(r->notes, "error-notes") == NULL)) {
		ap_table_setn(r->notes, "error-notes",
									ap_pvsprintf(r->pool, fmt, args));
	}
	va_end(args);
}


void ap_log_assert(const char *szExp, const char *szFile, int nLine)
{
	fprintf(stderr, "[%s] file %s, line %d, assertion \"%s\" failed\n",
					ap_get_time(), szFile, nLine, szExp);
#ifndef WIN32
	/* unix assert does an abort leading to a core dump */
	abort();
#else
	exit(1);
#endif
}
