/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDEING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */
/* xs.c - the main entry point for dreamweaver calls.
 *
 * original by alxander ladygo <alx@taylor.org>.
 */

/* on windows at least, uncomment this to write out 
 * debugging info to c:\tmp\debug.txt (dir must exist).
 */
/* #define FILE_DEBUG */

#include <stdio.h>
#include <io.h>

#ifdef macintosh 
#include <types.h>
#include <stat.h>
#else
#include <sys/types.h>
#include <sys/stat.h>
#endif

#include <fcntl.h>

#if defined(WIN32)
#include "os/win32/os.h"
#elif defined(macintosh)
#include "os.h"
#endif 

#include "mm_jsapi.h"
#include "httpd.h"
#include "alloc.h"
#include "buff.h"
#include "dw_strbuff.h"
#include "mod_include.h"
#include "enviro.h"
#include "xs.h"

#define ENV_ENTRIES 64

#ifdef WIN32
#define DEFAULT_DOCROOT "C:/"
#else
#define DEFAULT_DOCROOT "something mac-y"
#endif

char *get_docroot(pool *pool, const char *docNameStr);
char *get_filename(pool *pool, const char *docNameStr);
char *get_uri(pool *pool, const char *siteRootStr, char *docNameStr);
void add_env_vars(JSContext *cx, table *env, jsval vrs, jsval vls);
JSBool dw_build_rval(JSContext *cx, char *doc, char *errmsg, jsval *rval);

dw_strbuff *error_buff; /* errors are written here */

#ifdef FILE_DEBUG
FILE *debug;
#endif

/* codewarrior needs this for some reason.  probably my ignorance. -alx */
int main(void) { return 0; }

JSBool translate_markup(JSContext *cx, JSObject *obj,
												unsigned int argc,
												jsval *argv, jsval *rval)
{
	time_t now;
	pool *pool;
	request_rec *r;
	conn_rec *c;
	server_rec *s;
	dw_strbuff *sb;
	BUFF *buff;
	char *in, *docroot, *file_url;
	table *env;
	int success;
	
#ifdef FILE_DEBUG
	debug = fopen("/tmp/debug.txt", "a+");
	fputs("----hi\n", debug);
	fflush(debug);
#endif

	/* a couple things we'll need right off... */
	pool = ap_make_sub_pool(NULL);
	env = ap_make_table(pool, ENV_ENTRIES);
	
	/* setup error reporting buffer */
	if ((error_buff = (dw_strbuff *)ap_palloc(pool,	sizeof(dw_strbuff))) == NULL)
		return JS_FALSE;
	error_buff->outdata = NULL;
	error_buff->pool = pool;

	/* convert args */
	if (argc != 3 && argc != 5)
		return dw_build_rval(cx, "", "wrong number of args", rval);
	
	if ((file_url = JS_ValueToString(cx, argv[0], NULL)) == NULL)
		return dw_build_rval(cx, "", "converting docNameStr", rval);

	if ((docroot = JS_ValueToString(cx, argv[1], NULL)) == NULL)
		return dw_build_rval(cx, "", "converting siteRootStr", rval);

	if ((in = JS_ValueToString(cx, argv[2], NULL)) == NULL)
		return dw_build_rval(cx, "", "converting inStr", rval);

	
	/* memory allocation for structs & stuff */
	if ((r = ap_palloc(pool, sizeof(request_rec))) == NULL)
		return dw_build_rval(cx, "", "ap_palloc request_rec", rval);

	if ((c = (conn_rec *)ap_palloc(pool, sizeof(conn_rec))) == NULL)
		return dw_build_rval(cx, "", "ap_palloc conn_rec", rval);

	if ((s = (server_rec *)ap_palloc(pool, sizeof(server_rec))) == NULL)
		return dw_build_rval(cx, "", "ap_palloc server_rec", rval);

	if ((sb = (dw_strbuff *)ap_palloc(pool,	sizeof(dw_strbuff))) == NULL)
		return dw_build_rval(cx, "", "ap_palloc dw_strbuff", rval);


	if (setup_mod_include(pool))
		return dw_build_rval(cx, "", "setting up module", rval);
		
	/* setup request */
	time(&now);

	sb->pool = pool;
	sb->outdata = NULL;
	buff = ap_bcreate(pool, B_WR);
	buff->outdata = sb;
	c->client = buff;
	c->aborted = 0;

	s->ap_document_root = get_docroot(pool, docroot);
	s->path = "/";
	s->pathlen = 1;

	r->server = s;
	r->uri = get_uri(pool, s->ap_document_root, file_url);
	r->filename = get_filename(pool, file_url);
	r->pool = ap_make_sub_pool(pool);
	r->subprocess_env = env;
	r->notes = ap_make_table(r->pool, ENV_ENTRIES);
	r->headers_in = ap_make_table(r->pool, ENV_ENTRIES);
	r->connection = c;
	r->the_request = NULL;
	r->request_time = now;
	r->mtime = now;
	r->method = "GET";
	r->method_number = M_GET;
	r->path_info = NULL;
	r->file_data = in;
	r->main = NULL;
	r->args = NULL;

	/* this stuff would ordinarily be called, but we've circumvented
	 * it through our unconventional means of entry.  so just do it now. -alx */
	ap_add_common_vars(r);
	ap_add_cgi_vars(r);
	
	/* add these after the defaults, so we can overwrite */
	if (argc == 5)
		add_env_vars(cx, env, argv[3], argv[4]);
	
	send_parsed_content(NULL, r);	/* do it to it */

	if (ap_bflush(c->client))
		return dw_build_rval(cx, "", "flushing", rval);

#ifdef FILE_DEBUG
	fputs("writing:\n", debug);
	fputs(buff->outdata->outdata, debug);

	if (ap_table_get(r->notes, "error-notes") != NULL) {
		fputs("\nerror notes:\n", debug);
		fputs(ap_table_get(r->notes, "error-notes"), debug);
		if (NULL != error_buff->outdata) {
			fputs("\nerror buffer:\n", debug);
			fputs(error_buff->outdata, debug);
		}
	}
#endif /* FILE_DEBUG */
	
	success = dw_build_rval(cx, buff->outdata->outdata,
												 error_buff->outdata, rval);
	
	ap_destroy_pool(pool); 	/* god love apache pools */

#ifdef FILE_DEBUG
	fclose(debug);
#endif

	return success;
}

/* returns an array consisting of two arrays consisting of the keys
 * and values of several useful environental variables
 */
JSBool get_env(JSContext *cx, JSObject *obj, unsigned int argc,
							 jsval *argv, jsval *rval)
{
	request_rec *r;
	server_rec *s;
	pool *pool;
	table *env;
	char *file_url, *docroot;
	char *key, *val;
	jsval *keyVal, *valVal, keyval[2];
	JSObject *keyObj, *valObj;
	int i;
	
#ifdef FILE_DEBUG
	debug = fopen("/tmp/debug.txt", "a+");
	fputs("\n----env\n", debug);
	fflush(debug);
#endif

	pool = ap_make_sub_pool(NULL);
	env = ap_make_table(pool, ENV_ENTRIES);

	/* convert args */
	if (argc != 2)
		return dw_build_rval(cx, "", "wrong number of args", rval);
	
	if ((file_url = JS_ValueToString(cx, argv[0], NULL)) == NULL)
		return dw_build_rval(cx, "", "converting docNameStr", rval);

	if ((docroot = JS_ValueToString(cx, argv[1], NULL)) == NULL)
		return dw_build_rval(cx, "", "converting siteRootStr", rval);
	
	/* memory allocation for structs & stuff */
	if ((r = ap_palloc(pool, sizeof(request_rec))) == NULL)
		return dw_build_rval(cx, "", "ap_palloc request_rec", rval);

	if ((s = (server_rec *)ap_palloc(pool, sizeof(server_rec))) == NULL)
		return dw_build_rval(cx, "", "ap_palloc server_rec", rval);

	s->ap_document_root = get_docroot(pool, docroot);
	s->path = "/";
	s->pathlen = 1;

	r->server = s;
	r->uri = get_uri(pool, s->ap_document_root, file_url);
	r->filename = get_filename(pool, file_url);
	r->pool = ap_make_sub_pool(pool);
	r->subprocess_env = env;
	
	r->notes = ap_make_table(r->pool, ENV_ENTRIES);
	r->headers_in = ap_make_table(r->pool, ENV_ENTRIES);
	/*r->connection = c;*/
	r->the_request = NULL;
	/*r->request_time = now;
	r->mtime = now;*/
	r->method = "GET";
	r->method_number = M_GET;
	
	r->path_info = NULL;
	
	/*r->file_data = in;*/
	r->main = NULL;
	
	r->args = NULL;

	ap_add_common_vars(r);
	ap_add_cgi_vars(r);

	keyObj = JS_NewArrayObject(cx, ap_table_size(env), NULL);
	valObj = JS_NewArrayObject(cx, ap_table_size(env), NULL);

	/*
	if (NULL == keyval[0] || NULL == keyval[1]) {
		ap_destroy_pool(pool);
		return "creating arrays");
	}
	*/
	
	for (i = ap_table_size(env) - 1; i >= 0; --i) {
		key = ap_table_key_at(env, i);
		val = ap_table_get(env, key);
		if (!JS_StringToValue(cx, key, strlen(key), keyVal) ||
				!JS_StringToValue(cx, val, strlen(val), valVal))
		{
			ap_destroy_pool(pool);
			return dw_build_rval(cx, "", "converting values", rval);
		}
		JS_SetElement(cx, keyObj, i, keyVal);
		JS_SetElement(cx, valObj, i, valVal);
	}

	keyval[0] = JS_ObjectToValue(keyObj);
	keyval[1] = JS_ObjectToValue(valObj);
	*rval = JS_ObjectToValue(JS_NewArrayObject(cx, 2, keyval));

	ap_destroy_pool(pool);

#ifdef FILE_DEBUG
	fclose(debug);
#endif

	return JS_TRUE;
}


/* siteRootStr comes to us from dreamweaver in the form
 * "file:///C|/foo/".  we make mod_include happy by converting
 * to something of the form "c:/foo/".  -alx */
char *get_docroot(pool *pool, const char *siteRootStr)
{
	char *start, *root;

	start = strchr(siteRootStr, '/');

	if (NULL == start)
		return DEFAULT_DOCROOT;
	
	start += 3;

	if ((root = ap_palloc(pool, strlen(start) + 1)) == NULL) {
		dw_show_error("out of memory in get_docroot");
		return "";
	}

	strcpy(root, start);
#ifdef WIN32
	root[1] = ':';
#endif

#ifdef FILE_DEBUG
	fputs("\ndocroot: ", debug);
	fputs(root, debug);
	fputs("\n", debug);
#endif

	return root;
}

/* docNameStr comes to us from dreamweaver in the form
 * "file:///C|/foo/bar.html".  we make mod_include happy by converting
 * to something of the form "c:/foo/bar.html".  -alx */
char *get_filename(pool *pool, const char *docNameStr)
{
	char *start, *fname;

	start = strchr(docNameStr, '/');

	if (NULL == start)
		return "";

	start += 3;

	if ((fname = ap_palloc(pool, strlen(docNameStr) + 1)) == NULL) {
		dw_show_error("out of memory in get_filename");
		return "";
	}

	strcpy(fname, start);
#ifdef WIN32
	fname[1] = ':';
#endif

#ifdef FILE_DEBUG
	fputs("\nfname: ", debug);
	fputs(fname, debug);
	fputs("\n", debug);
#endif

	return fname;
}

/* docNameStr comes to us from dreamweaver in the form
 * "file:///C|/foo/bar.html" on windows, "file:///disk/foo/bar.html" on the mac.
 * the uri is the part after siteRootStr. 
 */
char *get_uri(pool *pool, const char *siteRootStr, char *docNameStr)
{
	char *start, *bar, *uri;

	bar = strchr(docNameStr, '|');
	if (NULL != bar)
		*bar = ':';
	
	start = strstr(docNameStr, siteRootStr);

	if (NULL == start)
		return "";

	/* siteRootStr must end in '/'.  we start with a '/'. */
	start += strlen(siteRootStr) - 1;
	
	if ((uri = ap_palloc(pool, strlen(docNameStr) + 1)) == NULL) {
		dw_show_error("out of memory in get_uri");
		return "";
	}

	strcpy(uri, start);

#ifdef FILE_DEBUG
	fputs("\nuri: ", debug);
	fputs(uri, debug);
	fputs("\n", debug);
#endif

	return uri;
}

/* take two passed in arrays, one of variable names, the other
 * their values, and add to our environmental vars.
 */
void add_env_vars(JSContext *cx, table *env, jsval vrs, jsval vls)
{
	JSObject *vars, *vals;
	jsval var, val;
	long l;
	char *varstr, *valstr;
	
	if (JS_ValueToObject(cx, vrs, &vars) == JS_FALSE) {
		dw_show_error("converting vars");
		return;
	}
	
	if (JS_ValueToObject(cx, vls, &vals) == JS_FALSE) {
		dw_show_error("converting vals");
		return;
	}

	if (vals == NULL || vars == NULL)
		return;

	if (JS_GetArrayLength(cx, vars) != JS_GetArrayLength(cx, vals)) {
		dw_show_error("vars.length != vals.length");
		return;
	}
		
	for (l = JS_GetArrayLength(cx, vars) - 1; l >= 0; --l) {
		JS_GetElement(cx, vals, (unsigned int)l, &val);
		JS_GetElement(cx, vars, (unsigned int)l, &var);
		varstr = JS_ValueToString(cx, var, NULL);
		valstr = JS_ValueToString(cx, val, NULL);
		ap_table_setn(env, varstr, valstr);
#ifdef FILE_DEBUG
		fputs("\nadded var: ", debug);
		fputs(varstr, debug);
		fputs(" = ", debug);
		fputs(valstr, debug);
#endif
	}
}


JSBool dw_show_error(char *msg)
{
	dw_strwrite(error_buff, msg, strlen(msg));
	return JS_FALSE;
}


JSBool dw_build_rval(JSContext *cx, char *doc, char *errmsg, jsval *rval)
{
	int success, i;
	jsval outVal[2];
	
#ifdef macintosh
	/* handle \n -> \r translation on the mac */
	for (i = strlen(errmsg) - 1; i >= 0; --i)
		if (errmsg[i] == '\n')
			errmsg[i] = '\r';
#endif

#ifdef JS_REPORT_ERROR_WORKS

	/* currently not the case.  this is how things should work if
	 * JS_ReportError ever decides to get fixed.
	 */
	if (errmsg == NULL) {
		success = JS_StringToValue(cx, doc, 0, rval);
	} else {
		JS_StringToValue(cx, "", 0, rval);
		success = JS_FALSE;
		JS_ReportError(cx, errmsg, 0);
	}
	
#else /* JS_REPORT_ERROR_WORKS */

	/* grrr.  we return JS_TRUE to the translator, along with an
	 * array containing the translated doc and the error message.  either
	 * can be the empty string.  if the doc is empty, an error has occurred,
	 * and should be displayed to the user.
	 */
	success = JS_TRUE;
	if (errmsg == NULL) {
		JS_StringToValue(cx, doc, 0, &outVal[0]);
		JS_StringToValue(cx, "", 0, &outVal[1]);
	} else {
		JS_StringToValue(cx, "", 0, &outVal[0]);
		JS_StringToValue(cx, errmsg, 0, &outVal[1]);
	}

	*rval = JS_ObjectToValue(JS_NewArrayObject(cx, 2, outVal));
	
#endif /* JS_REPORT_ERROR_WORKS */

	return success;
}

/* stuff required to link the dll to dreamweaver. */
MM_STATE

void MM_Init()
{
	JS_DefineFunction("translate_markup", translate_markup, 3);
	JS_DefineFunction("get_env", get_env, 2);
}
